"""
    This script is used to train a Conformer model for conformer prediction.
"""

import os, pdb, pickle, json
import wandb
import warnings
import torch
from torch_geometric.utils import degree
from argparse import Namespace
from datasets import load_from_disk, DatasetDict, concatenate_datasets
from molecule3d import HFMolecule3DRandomSplit, HFMolecule3DScaffoldSplit, HFQm9
from dataclasses import dataclass

from transformers import EvalPrediction
from transformers import TrainingArguments, Trainer
from transformers import HfArgumentParser
from transformers.trainer_utils import get_last_checkpoint

from models.ours import Config, OursConformerPrediction, Collator
import torch.multiprocessing as mp
import multiprocessing
import torch.nn.functional as F
import numpy as np
import random


device = "cuda" if torch.cuda.is_available() else "cpu"
warnings.simplefilter("ignore", UserWarning)

def init_seed(seed):
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    random.seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False

@dataclass
class ScriptArguments:
    tokenized_dataset_path: str
    dataset: str
    num_layers: int
    split: str
    edge_type: str
    k: int


def compute_metrics(eval_pred: EvalPrediction):
    # EvalPrediction(predictions=all_preds, label_ids=all_labels) from transformers Trainer with args.include_inputs_for_metrics=False.
    preds, _ = eval_pred
    mae, mse, rmsd, *_ = preds
    mae, mse, rmsd = mae.mean().item(), mse.mean().item(), rmsd.mean().item()

    # print(f"MAE: {mae:.3f}, RMSE: {rmse:.3f}")
    return {"mae": mae, "mse": mse, "rmsd": rmsd}


def main():
    parse = HfArgumentParser(dataclass_types=[ScriptArguments, TrainingArguments])
    script_args, training_args = parse.parse_args_into_dataclasses()
    print(script_args)
    print(training_args)
    
    init_seed(training_args.seed)

    if script_args.dataset == "QM9":
        dataset = DatasetDict.load_from_disk(
            "/mnt/ssd/$$$$$$$$$/" + script_args.tokenized_dataset_path + "_prox")
    elif script_args.dataset == "Molecule3D":
        if script_args.split == "random":
            path = os.path.join("/mnt/ssd/$$$$$$$$$", script_args.tokenized_dataset_path, script_args.split)
        elif script_args.split == "scaffold":
            path = os.path.join("/mnt/ssd/$$$$$$$$$", script_args.tokenized_dataset_path,
                                script_args.split + "_ete")
        dataset = DatasetDict.load_from_disk(path)

    train_set, eval_set, test_set = (
        dataset["train"],
        dataset["validation"],
        dataset["test"],
    )
    
    print("Laplacian K: ", script_args.k)
    collate_func = Collator(
        edge_type=script_args.edge_type, 
        laplacian_k=script_args.k,
    )

    if script_args.edge_type == "default":
        embed_style = "atom_tokenized_ids"
    else:
        embed_style = "atom_type_ids"

    print("Embed style: ", embed_style)
    config = Config(
        # encoder config
        n_encode_layers=script_args.num_layers,
        encoder_use_A_in_attn=True if script_args.edge_type == "default" else False,
        encoder_use_D_in_attn=False,
        # decoder config
        n_decode_layers=script_args.num_layers,
        decoder_use_A_in_attn=True if script_args.edge_type == "default" else False,
        decoder_use_D_in_attn=False,
        embed_style=embed_style,
        atom_vocab_size=513,  # 51 2 + 1 for padding (shifted all input ids by 1)
        d_embed=256 if script_args.edge_type == "default" else 512,
        pre_ln=False,  # layer norm before residual, else after residual, not Pre-LN and not Post-LN.
        d_q=256 if script_args.edge_type == "default" else 512,
        d_k=256 if script_args.edge_type == "default" else 512,
        d_v=256 if script_args.edge_type == "default" else 512,
        d_model=256 if script_args.edge_type == "default" else 512,
        n_head=8,
        qkv_bias=True,
        attn_drop=0.,
        norm_drop=0.,
        ffn_drop=0.,
        dropout=0.,
        d_ffn=1024,
        k=script_args.k,
        edge_type=script_args.edge_type
    )

    model = OursConformerPrediction(config)
    print(model.config)

    trainer = Trainer(
        model=model,
        args=training_args,
        data_collator=collate_func,
        train_dataset=train_set,
        eval_dataset=eval_set,
        compute_metrics=compute_metrics,
    )
    
    # training
    if training_args.do_train:
        resume_from_checkpoint = training_args.resume_from_checkpoint
        try:
            resume_f = eval(resume_from_checkpoint)
            if not resume_f:
                train_result = trainer.train()
            else:
                last_checkpoint = get_last_checkpoint(training_args.output_dir)
                train_result = trainer.train(resume_from_checkpoint=last_checkpoint)
        except Exception:
            train_result = trainer.train()
        trainer.log_metrics("train", train_result.metrics)


    # testing
    if training_args.do_eval:
        # TODO: if training_args.do_train is False, model used to test is not the best!
        test_metrics = trainer.evaluate(eval_dataset=test_set)
        trainer.log_metrics("test", test_metrics)

if __name__ == "__main__":
    main()
