"""
    This module contains basic GNN modules used in MoleBert.
"""

from typing import Any
import torch
from torch import Tensor
import torch.nn as nn
import torch.nn.functional as F
from torch_geometric.nn import MessagePassing
from torch_geometric.utils import add_self_loops, remove_self_loops, degree

import math
from torch_geometric.utils import softmax
from typing import Optional, Union, Tuple, List, Dict, Callable
from torch_geometric.nn import BatchNorm
from torch_geometric.nn.conv import PNAConv
from torch_geometric.typing import OptTensor, Adj, PairTensor, OptPairTensor
from torch_geometric.nn.dense.linear import Linear
from torch_geometric.nn.inits import reset, zeros
from copy import deepcopy
from collections import OrderedDict
from torch_geometric.nn.inits import glorot, zeros, reset, ones
from torch.nn import Parameter
from torch_sparse import SparseTensor
from torch_geometric.utils import spmm, to_dense_batch


NUM_ATOM_TYPE = 120  # Atom numbers 1~118 + extra mask tokens   TODO: What is the extra mask token?
NUM_CHIRALITY_TAG = 3  # {UNSPECIFIED, TETRAHEDRAL_CW, CHI_TETRAHEDRAL_CCW, CHI_OTHER}     TODO: Why 3?
NUM_BOND_TYPE = 6  # {SINGLE, DOUBLE, TRIPLE, AROMATIC, SELF_LOOP} + masked tokens    TODO: What is masked tokens?
NUM_BOND_DIRECTION = 3  # {NONE, ENDUPRIGHT, ENDDOWNRIGHT}


class HANConv(MessagePassing):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        num_relations: int,
        heads: int = 1,
        negative_slope=0.2,
        dropout: float = 0.0,
        edge_dim: Optional[int] = None,
        add_self_loops: bool = True,
        fill_value: Union[float, Tensor, str] = 'mean',
        **kwargs,
    ):
        super().__init__(aggr='add', node_dim=0, **kwargs)
        self.heads = heads
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.negative_slope = negative_slope
        self.dropout = dropout
        self.edge_dim = edge_dim
        self.add_self_loops = add_self_loops
        self.fill_value = fill_value
        ### For semantic attention ###
        if num_relations > 1:
            self.k_lin = nn.Linear(out_channels, out_channels)
            self.q = nn.Parameter(torch.empty(1, out_channels))

        self.proj = Linear(in_channels, out_channels, bias=False,
                                   weight_initializer='glorot')

        ### For node-level attention ###
        self.lin_src = nn.ParameterDict()
        self.lin_dst = nn.ParameterDict()
        dim = out_channels // heads
        for et in range(num_relations):
            self.lin_src[str(et)] = nn.Parameter(torch.empty(1, heads, dim))
            self.lin_dst[str(et)] = nn.Parameter(torch.empty(1, heads, dim))

        if edge_dim is not None:
            self.proj_edge = Linear(edge_dim, out_channels, bias=False,
                                   weight_initializer='glorot')
            self.lin_edge = nn.ParameterDict()
            for et in range(num_relations):
                self.lin_edge[str(et)] = nn.Parameter(torch.empty(1, heads, dim))

        self.reset_parameters()

    def reset_parameters(self):
        reset(self.proj)
        glorot(self.lin_src)
        glorot(self.lin_dst)
        self.k_lin.reset_parameters()
        glorot(self.q)
        if self.edge_dim is not None:
            reset(self.proj_edge)
            glorot(self.lin_edge)

    def forward(
        self,
        x: Union[Tensor, List[Tensor]],
        edge_index_list: Union[Adj, List[Adj]],
        edge_attr_list: Union[OptTensor, List[OptTensor]] = None
    ):
        H, D = self.heads, self.out_channels // self.heads
        x_lin = self.proj(x).view(-1, H, D)
        out_list = []

        # Node-level attention within a single edge type (same as GAT)
        for et, edge_index in enumerate(edge_index_list):
            
            if self.add_self_loops:
                num_nodes = x_lin.size(0)
                if edge_attr_list is not None:
                    edge_index, edge_attr = remove_self_loops(
                        edge_index, edge_attr_list[et])
                else:
                    edge_index, edge_attr = remove_self_loops(
                        edge_index, None)
                edge_index, edge_attr = add_self_loops(
                    edge_index, edge_attr, fill_value=self.fill_value,
                    num_nodes=num_nodes)

            x_src = x_dst = x_lin
            if edge_attr is not None and self.edge_dim is not None:
                edge_lin = self.proj_edge(edge_attr).view(-1, H, D)
                edge_lin = (edge_lin * self.lin_edge[str(et)]).sum(dim=-1)
            else:
                edge_lin = None
            
            lin_src = self.lin_src[str(et)]
            lin_dst = self.lin_dst[str(et)]
            alpha_src = (x_src * lin_src).sum(dim=-1) # (N, H)
            alpha_dst = (x_dst * lin_dst).sum(dim=-1)
            # node-level attention
            out = self.propagate(edge_index, x=(x_src, x_dst),
                                 alpha=(alpha_src, alpha_dst),
                                 edge_attr=edge_lin) # (N, H*D=out_channels)
            out = F.relu(out)
            out_list.append(out)
        
        # type-level attention
        if len(edge_index_list) > 1:
            out = torch.stack(out_list) # (E, N, out_channels)
            sem_attn = (self.q * torch.tanh(self.k_lin(out)).mean(1)).sum(-1) # (E,)
            sem_attn = F.softmax(sem_attn, dim=0)
            out = torch.sum(sem_attn.view(len(edge_index_list), 1, -1) * out, dim=0) # (N, out_channels)
            # print(sem_attn)
        return out

    def message(self, x_j: Tensor,
                alpha_i: Tensor, alpha_j: Tensor,
                edge_attr: OptTensor, index: Tensor, 
                ptr: Optional[Tensor],
                size_i: Optional[int]) -> Tensor:

        alpha = alpha_j + alpha_i
        if edge_attr is not None:
            alpha = alpha + edge_attr
        alpha = F.leaky_relu(alpha, self.negative_slope)
        alpha = softmax(alpha, index, ptr, size_i)
        alpha = F.dropout(alpha, p=self.dropout, training=self.training)
        out = x_j * alpha.view(-1, self.heads, 1)
        return out.view(-1, self.out_channels)

    def __repr__(self) -> str:
        return (f'{self.__class__.__name__}({self.out_channels}, '
                f'heads={self.heads})')


class RGCNConv(MessagePassing):
    def __init__(self, in_channels: Union[int, Tuple[int, int]],
                 out_channels: int,
                 num_relations: int,
                 num_bases: Optional[int] = None,
                 aggr: str = 'mean',
                 root_weight: bool = False,
                 bias: bool = True, 
                 edge_dim: Optional[int] = None,
                 add_self_loops: bool = True,
                 fill_value: Union[float, Tensor, str] = 'mean',
                 dropout: float = 0.,
                 **kwargs
    ):  # yapf: disable

        super(RGCNConv, self).__init__(aggr=aggr, node_dim=0, **kwargs)

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.num_relations = num_relations
        self.num_bases = num_bases
        self.edge_dim = edge_dim
        self.add_self_loops = add_self_loops
        self.fill_value = fill_value
        self.dropout = dropout
        
        if isinstance(in_channels, int):
            in_channels = (in_channels, in_channels)
        self.in_channels_l = in_channels[0]

        if num_bases is not None:
            self.weight = Parameter(
                torch.Tensor(num_bases, in_channels[0], out_channels))
            self.comp = nn.ParameterDict()
            for r in range(num_relations):
                self.comp[str(r)] = Parameter(torch.Tensor(1, num_bases))
        else:
            self.weight = nn.ParameterDict()
            for r in range(num_relations):
                self.weight[str(r)] = Parameter(
                    torch.Tensor(in_channels[0], out_channels))
            self.register_parameter('comp', None)

        if root_weight:
            self.root = Parameter(torch.Tensor(in_channels[1], out_channels))
        else:
            self.register_parameter('root', None)

        if bias:
            self.bias = Parameter(torch.Tensor(out_channels))
        else:
            self.register_parameter('bias', None)
        
        if num_relations > 1:
            self.weights_rel = Linear(out_channels, 1, bias=True,
                                      weight_initializer='glorot')

        if edge_dim is not None:
            self.proj_edge = Parameter(
                torch.Tensor(num_relations, edge_dim, out_channels))
        self.reset_parameters()

    def reset_parameters(self):
        glorot(self.weight)
        glorot(self.comp)
        glorot(self.root)
        zeros(self.bias)
        if self.num_relations > 1:
            reset(self.weights_rel)
        if self.edge_dim is not None:
            glorot(self.proj_edge)

    def forward(
            self, 
            x: Union[OptTensor, Tuple[OptTensor, Tensor]],
            edge_index_list: Union[Adj, List[Adj]], 
            edge_attr_list: Union[OptTensor, List[OptTensor]] = None
        ):
        
        x_l: OptTensor = None
        if isinstance(x, tuple):
            x_l = x[0]
        else:
            x_l = x
        if x_l is None:
            x_l = torch.arange(self.in_channels_l, device=self.weight.device)

        x_r: Tensor = x_l
        if isinstance(x, tuple):
            x_r = x[1]

        size = (x_l.size(0), x_r.size(0))

        # propagate_type: (x: Tensor)
        out = torch.zeros(x_r.size(0), self.out_channels, self.num_relations, device=x_r.device)

        empty_relation = [r for r in range(self.num_relations) if not edge_index_list[r].numel()]
        if self.num_bases is not None:  # Basis-decomposition =================
            weight = self.weight
            weight_list = []
            for r in range(self.num_relations):
                if r not in empty_relation:
                    # Perform the matrix multiplication and reshaping for each relation
                    w_r = self.comp[str(r)] @ weight.view(self.num_bases, -1)
                    w_r = w_r.view(self.in_channels_l, self.out_channels)
                    weight_list.append(w_r)
            
            weight = torch.stack(weight_list, dim=0)

        for r in range(self.num_relations):
            if r not in empty_relation:
                num_nodes=x_l.size(0)
                edge_index = edge_index_list[r]
                if self.add_self_loops:
                    if edge_attr_list is not None:
                        edge_index, edge_attr = remove_self_loops(
                            edge_index, edge_attr_list[r])
                    else:
                        edge_index, edge_attr = remove_self_loops(
                            edge_index, None)
                    edge_index, edge_attr = add_self_loops(
                        edge_index, edge_attr, fill_value=self.fill_value,
                        num_nodes=num_nodes)
                    
                if edge_attr is not None and self.edge_dim is not None:
                    edge_attr = edge_attr @ self.proj_edge[r]
                else:
                    edge_attr = None

                if x_l.dtype == torch.long:
                    if self.num_bases is not None:
                        out[:, :, r] = self.propagate(edge_index, x=weight[r, x_l], edge_attr=edge_attr, size=size) 
                    else:
                        out[:, :, r] = self.propagate(edge_index, self.weight[str(r)][x_l], edge_attr=edge_attr, size=size)
                else:
                    h = self.propagate(edge_index, x=x_l, edge_attr=edge_attr, size=size)
                    if self.num_bases is not None:
                        out[:, :, r] = (h @ weight[r])
                    else:
                        out[:, :, r] = (h @ self.weight[str(r)])

        if self.num_relations > 1:
            attn_scores = self.weights_rel(out.transpose(1,2)).squeeze(-1) # (N, num_rel)
            scale = out.shape[1] ** 0.5
            attn_scores = attn_scores / torch.tensor(scale)
            attn_weights = torch.softmax(attn_scores, dim=-1).unsqueeze(1)
            attn_weights = F.dropout(attn_weights, p=self.dropout, training=self.training)
            out = torch.sum(out * attn_weights, dim=-1) # (N, out_channels)
        else:
            out = out.squeeze(-1)
            
        if self.bias is not None:
            out += self.bias

        return out

    def message(self, x_j: Tensor, edge_attr: OptTensor) -> Tensor:
        if edge_attr is not None:
            return x_j  + edge_attr
        else:
            return x_j

    def __repr__(self):
        return '{}({}, {}, num_relations={})'.format(self.__class__.__name__,
                                                     self.in_channels,
                                                     self.out_channels,
                                                     self.num_relations)


class RGCNTransformerConv(MessagePassing):
    def __init__(self, in_channels: Union[int, Tuple[int, int]],
                 out_channels: int,
                 num_relations: int,
                 num_bases: Optional[int] = None,
                 aggr: str = 'mean',
                 bias: bool = True, 
                 edge_dim: Optional[int] = None,
                 add_self_loops: bool = True,
                 fill_value: Union[float, Tensor, str] = 'mean',
                 dropout: float = 0.,
                 **kwargs
    ):  # yapf: disable

        super(RGCNTransformerConv, self).__init__(aggr=aggr, node_dim=0, **kwargs)

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.num_relations = num_relations
        self.num_bases = num_bases
        self.edge_dim = edge_dim
        self.add_self_loops = add_self_loops
        self.fill_value = fill_value
        self.dropout = dropout
        
        if isinstance(in_channels, int):
            in_channels = (in_channels, in_channels)
        self.in_channels_l = in_channels[0]

        self.mha = nn.MultiheadAttention(
                in_channels[0],
                8,
                batch_first=True
        )

        if num_bases is not None:
            self.weight = Parameter(
                torch.Tensor(num_bases, in_channels[0], out_channels))
            self.comp = nn.ParameterDict()
            for r in range(num_relations):
                self.comp[str(r)] = Parameter(torch.Tensor(1, num_bases))
        else:
            self.weight = nn.ParameterDict()
            for r in range(num_relations):
                self.weight[str(r)] = Parameter(
                    torch.Tensor(in_channels[0], out_channels))
            self.register_parameter('comp', None)

        if bias:
            self.bias = Parameter(torch.Tensor(out_channels))
        else:
            self.register_parameter('bias', None)
        
        if num_relations > 1:
            self.weights_rel = Linear(out_channels, 1, bias=True,
                                      weight_initializer='glorot')

        if edge_dim is not None:
            self.proj_edge = Parameter(
                torch.Tensor(num_relations, edge_dim, out_channels))
        self.reset_parameters()

    def reset_parameters(self):
        glorot(self.weight)
        glorot(self.comp)
        zeros(self.bias)
        self.mha._reset_parameters()
        if self.num_relations > 1:
            reset(self.weights_rel)
        if self.edge_dim is not None:
            glorot(self.proj_edge)

    def forward(
            self, 
            x: Union[OptTensor, Tuple[OptTensor, Tensor]],
            edge_index_list: Union[Adj, List[Adj]], 
            edge_attr_list: Union[OptTensor, List[OptTensor]] = None,
            batch: Tensor = None
        ):
        
        # propagate_type: (x: Tensor)
        ## global
        h_g, mask = to_dense_batch(x, batch)
        h_g, _ = self.mha(h_g, h_g, h_g, key_padding_mask=~mask,
                             need_weights=False)
        h_g = h_g[mask]

        h_l = torch.zeros(x.size(0), self.out_channels, self.num_relations+1, device=x.device)
        h_l[:, :, 0] = h_g
        empty_relation = [r for r in range(self.num_relations) if not edge_index_list[r].numel()]
        if self.num_bases is not None:  # Basis-decomposition =================
            weight = self.weight
            weight_list = []
            for r in range(self.num_relations):
                if r not in empty_relation:
                    # Perform the matrix multiplication and reshaping for each relation
                    w_r = self.comp[str(r)] @ weight.view(self.num_bases, -1)
                    w_r = w_r.view(self.in_channels_l, self.out_channels)
                    weight_list.append(w_r)
            
            # Concatenate the weights for all non-empty relations
            weight = torch.stack(weight_list, dim=0)  # Stack instead of cat to keep the relation dimension

        for r in range(self.num_relations):
            if r not in empty_relation:
                num_nodes=x.size(0)
                edge_index = edge_index_list[r]
                if self.add_self_loops:
                    if edge_attr_list is not None:
                        edge_index, edge_attr = remove_self_loops(
                            edge_index, edge_attr_list[r])
                    else:
                        edge_index, edge_attr = remove_self_loops(
                            edge_index, None)
                    edge_index, edge_attr = add_self_loops(
                        edge_index, edge_attr, fill_value=self.fill_value,
                        num_nodes=num_nodes)
                    
                if edge_attr is not None and self.edge_dim is not None:
                    edge_attr = edge_attr @ self.proj_edge[r]
                else:
                    edge_attr = None

                h = self.propagate(edge_index, x=x, edge_attr=edge_attr)
                if self.num_bases is not None:
                    h_l[:, :, r+1] = (h @ weight[r])
                else:
                    h_l[:, :, r+1] = (h @ self.weight[str(r)])

        if self.num_relations > 1:
            attn_scores = self.weights_rel(h_l.transpose(1,2)).squeeze(-1) # (N, num_rel)
            scale = h_l.shape[1] ** 0.5
            attn_scores = attn_scores / torch.tensor(scale)
            attn_weights = torch.softmax(attn_scores, dim=-1).unsqueeze(1)
            attn_weights = F.dropout(attn_weights, p=self.dropout, training=self.training)
            h_l = torch.sum(h_l * attn_weights, dim=-1) # (N, out_channels)
        else:
            h_l = h_l.squeeze(-1)
            
        if self.bias is not None:
            h_l += self.bias

        # out = h_g + h_l
        # out_dict = {"global": h_g, "local": h_l, "total": out}
        out_dict = {"total": h_l}
        return out_dict

    def message(self, x_j: Tensor, edge_attr: OptTensor) -> Tensor:
        if edge_attr is not None:
            return x_j  + edge_attr
        else:
            return x_j

    def __repr__(self):
        return '{}({}, {}, num_relations={})'.format(self.__class__.__name__,
                                                     self.in_channels,
                                                     self.out_channels,
                                                     self.num_relations)


class GINConv(MessagePassing):
    """
    Extension of GIN aggregation to incorporate edge information by concatenation.
    """

    def __init__(self, in_dim: int, hidden_dim: int, out_dim: int, aggr: str = "add", **kwargs):
        kwargs.setdefault("aggr", aggr)
        super().__init__(**kwargs)
        self.aggr = aggr
        self.mlp = nn.Sequential(nn.Linear(in_dim, hidden_dim), nn.ReLU(), nn.Linear(hidden_dim, out_dim))
        self.edge_embedding01 = nn.Embedding(NUM_BOND_TYPE, in_dim)
        self.edge_embedding02 = nn.Embedding(NUM_BOND_DIRECTION, in_dim)

    def forward(self, **inputs) -> Any:
        node_attr = inputs.get("node_attr")
        edge_index = inputs.get("edge_index")
        edge_type = inputs.get("edge_type")
        edge_dire_type = inputs.get("edge_dire_type")
        # add self loops in the edge space
        edge_index, _ = add_self_loops(edge_index, num_nodes=node_attr.shape[0])
        # edge_index.shape = [2, num_edges + num_nodes]
        # [
        #   [0, 1, 2, 3, 4, 0, 1, 2, 3, 4],
        #   [1, 2, 3, 4, 0, 0, 1, 2, 3, 4]
        # ]

        # add features corresponding to self-loop edges
        # bond direction : NONE   bond type : SELF_LOOP
        self_loop_attr = torch.zeros(node_attr.size(0), 2, dtype=torch.long, device=node_attr.device)
        self_loop_attr[:, 0] = 4
        # computing edge embedding: bond type embedded feature + bond direction embedded feature
        edge_attr = self.edge_embedding01(edge_type) + self.edge_embedding02(edge_dire_type)
        self_loop_attr = self.edge_embedding01(self_loop_attr[:, 0]) + self.edge_embedding02(self_loop_attr[:, 1])
        edge_attr = torch.cat([edge_attr, self_loop_attr], dim=0)
        # edge_attr.shape = [num_edges + num_nodes, emb_dim]

        # propagate():
        # 1. message(): aggregate messages from neighbors;
        # 2. aggregate(): combine messages with self embeddings;
        # 3. update(): update self embeddings with aggregated messages.
        h = self.propagate(edge_index, x=node_attr, edge_attr=edge_attr)
        return h

    def message(self, x_j: torch.Tensor, edge_attr: torch.Tensor) -> torch.Tensor:
        # x_j.shape = [num_edges, emb_dim]  TODO: What is x_j and How to get it?
        # TODO: Is this edge_attr the same as the edge_attr in forward()?
        return x_j + edge_attr

    def update(self, aggr_out: torch.Tensor) -> torch.Tensor:
        # aggr_out.shape = [num_nodes, out_dim]
        return self.mlp(aggr_out)


class GNNEncoder(nn.Module):
    def __init__(self, num_layers: int = 5, embedding_dim=300, layer_hidden_dim: int = 600, jk="last", dropout: float = 0.0, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.num_layers = num_layers
        self.embedding_dim = embedding_dim
        self.jk = jk
        self.dropout = dropout

        if self.num_layers < 2:
            raise ValueError("Number of GNN layers must be greater than 1.")

        self.node_embedding01 = nn.Embedding(NUM_ATOM_TYPE, embedding_dim)
        self.node_embedding02 = nn.Embedding(NUM_CHIRALITY_TAG, embedding_dim)

        self.gnn_layers = nn.ModuleList()
        for _ in range(self.num_layers):
            self.gnn_layers.append(GINConv(embedding_dim, layer_hidden_dim, embedding_dim))

        self.batch_norms = nn.ModuleList()
        for _ in range(self.num_layers):
            self.batch_norms.append(nn.BatchNorm1d(embedding_dim))

    def forward(self, **inputs):
        node_type = inputs.get("node_type")
        node_chiral_type = inputs.get("node_chiral_type")
        edge_type = inputs.get("edge_type")
        edge_dire_type = inputs.get("edge_dire_type")
        edge_index = inputs.get("edge_index")

        node_chiral_type[node_chiral_type>=3] = 2

        # computing input node embedding: atom type embedded feature + chirality embedded feature
        node_attr = self.node_embedding01(node_type) + self.node_embedding02(node_chiral_type)
        h_list = [node_attr]  # list of hidden representation at each layer (including input)

        for i in range(self.num_layers):
            h = self.gnn_layers[i](node_attr=h_list[i], edge_index=edge_index, edge_type=edge_type, edge_dire_type=edge_dire_type)
            h = self.batch_norms[i](h)
            # h.shape = (num_nodes, emb_dim)
            if i == self.num_layers - 1:
                # remove relu activation for the last layer
                h = F.dropout(h, self.dropout, training=self.training)
            else:
                h = F.dropout(F.relu(h), self.dropout, training=self.training)
            h_list.append(h)

        # different implementations of Jk-concat
        if self.jk == "concat":
            node_representation = torch.cat(h_list, dim=1)
            # node_representation.shape = (num_nodes, emb_dim * num_layer)
        elif self.jk == "last":
            node_representation = h_list[-1]
            # node_representation.shape = (num_nodes, emb_dim)
        elif self.jk == "max":  # position-wise max
            h_list = [h.unsqueeze(0) for h in h_list]  # h.shape = (1, num_nodes, emb_dim)
            node_representation = torch.max(torch.cat(h_list, dim=0), dim=0)[0]
            # node_representation.shape = (num_nodes, emb_dim)
        elif self.jk == "sum":  # position-wise sum
            h_list = [h.unsqueeze(0) for h in h_list]
            node_representation = torch.sum(torch.cat(h_list, dim=0), dim=0)[0]
            # node_representation.shape = (num_nodes, emb_dim)

        return node_representation  # [num_nodes, emb_dim] / [num_nodes, emb_dim * num_layer]

    def __init__weights__(self):
        for m in self.modules():
            if isinstance(m, nn.Embedding):
                nn.init.xavier_uniform_(m.weight.data)
            elif isinstance(m, nn.Linear):
                nn.init.xavier_uniform_(m.weight.data)
                if m.bias is not None:
                    nn.init.zeros_(m.bias.data)


class GNNDecoder(nn.Module):
    def __init__(self, in_dim: int = 300, hidden_dim: int = 600, out_dim: int = 300, dropout: float = 0.0):
        super().__init__()
        self.dropout = dropout
        self.conv = GINConv(in_dim, hidden_dim, out_dim)
        self.act = nn.PReLU()
        self.enc_to_dec = nn.Linear(in_dim, in_dim, bias=False)

    def forward(self, **inputs):
        node_attr = inputs.get("node_attr")
        edge_index = inputs.get("edge_index")
        edge_type = inputs.get("edge_type")
        edge_dire_type = inputs.get("edge_dire_type")
        node_attr = self.act(node_attr)
        node_attr = self.enc_to_dec(node_attr)
        node_attr = F.dropout(node_attr, self.dropout, training=self.training)
        out = self.conv(node_attr=node_attr, edge_index=edge_index, edge_type=edge_type, edge_dire_type=edge_dire_type)
        return out


class TransformerConv(MessagePassing):
    _alpha: OptTensor

    def __init__(
        self,
        in_channels: Union[int, Tuple[int, int]],
        out_channels: int,
        heads: int = 1,
        concat: bool = True,
        beta: bool = False,
        dropout: float = 0.,
        bias: bool = True,
        root_weight: bool = True,
        edge_dim = None,
        **kwargs,
    ):
        kwargs.setdefault('aggr', 'add')
        super(TransformerConv, self).__init__(node_dim=0, **kwargs)

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.heads = heads
        self.beta = beta and root_weight
        self.root_weight = root_weight
        self.concat = concat
        self.dropout = dropout
        if edge_dim is not None:
            self.edge_dim = True
        else:
            self.edge_dim = False

        if isinstance(in_channels, int):
            in_channels = (in_channels, in_channels)

        self.lin_key = Linear(in_channels[0], heads * out_channels)
        self.lin_query = Linear(in_channels[1], heads * out_channels)
        self.lin_value = Linear(in_channels[0], heads * out_channels)
        if edge_dim is not None:
            self.lin_edge = Linear(edge_dim, heads * out_channels, bias=False)
        else:
            self.lin_edge = self.register_parameter('lin_edge', None)

        if concat:
            self.lin_skip = Linear(in_channels[1], heads * out_channels,
                                   bias=bias)
            self.lin_ema = Linear(3 * heads * out_channels, 1, bias=False)

        else:
            self.lin_skip = Linear(in_channels[1], out_channels, bias=bias)
            self.lin_ema = Linear(3 * out_channels, 1, bias=False)
            
        self.reset_parameters()

    def reset_parameters(self):
        self.lin_key.reset_parameters()
        self.lin_query.reset_parameters()
        self.lin_value.reset_parameters()
        if self.edge_dim:
            self.lin_edge.reset_parameters()
        self.lin_skip.reset_parameters()
        self.lin_ema.reset_parameters()

    def forward(self, x: Union[Tensor, PairTensor], edge_index: Adj, 
                edge_attr: OptTensor = None, return_attention_weights=None):
        if isinstance(x, Tensor):
            x: PairTensor = (x, x)

        out = self.propagate(edge_index, x=x, edge_attr=edge_attr, size=None)
        alpha = self._alpha
        self._alpha = None

        if self.concat:
            out = out.view(-1, self.heads * self.out_channels)
        else:
            out = out.mean(dim=1)

        if self.root_weight:
            x_r = self.lin_skip(x[1])
            beta = self.lin_ema(torch.cat([out, x_r, out - x_r], dim=-1))
            beta = beta.sigmoid()
            out = beta * x_r + (1 - beta) * out

        return out

    def message(self, x_i: Tensor, x_j: Tensor, edge_attr: OptTensor,
                index: Tensor, ptr: OptTensor,
                size_i: Optional[int]) -> Tensor:

        query = self.lin_query(x_i).view(-1, self.heads, self.out_channels)
        key = self.lin_key(x_j).view(-1, self.heads, self.out_channels)
        if self.edge_dim:
            assert edge_attr is not None
            edge_attr = self.lin_edge(edge_attr).view(-1, self.heads,
                                                      self.out_channels)
            key += edge_attr
        alpha = (query * key).sum(dim=-1) / math.sqrt(self.out_channels)
        alpha = softmax(alpha, index, ptr, size_i)
        self._alpha = alpha
        alpha = F.dropout(alpha, p=self.dropout, training=self.training)

        out = self.lin_value(x_j).view(-1, self.heads, self.out_channels)
        if self.edge_dim:
            out += edge_attr

        out *= alpha.view(-1, self.heads, 1)
        return out

    def __repr__(self):
        return '{}({}, {}, heads={})'.format(self.__class__.__name__,
                                             self.in_channels,
                                             self.out_channels, self.heads)


class UniMP(nn.Module):
    def __init__(self, num_layers, in_channels, num_hidden,
                 drop_rate, num_heads=8, edge_dim=None):
        super(UniMP, self).__init__()

        self.num_layers = num_layers
        self.num_heads = num_heads
        self.edge_dim = edge_dim
        
        self.convs = torch.nn.ModuleList()
        self.norms = torch.nn.ModuleList()
        for i in range(num_layers):
            self.convs.append(
                TransformerConv(num_hidden, num_hidden//num_heads,
                                heads = num_heads,
                                concat = True,
                                dropout = drop_rate,
                                beta = True,
                                edge_dim = edge_dim,
                                )
                            )
            if i != num_layers - 1:
                self.norms.append(nn.LayerNorm(num_hidden))
        self.convs[0] = TransformerConv(in_channels, num_hidden//num_heads,
                                        heads = num_heads,
                                        concat = True,
                                        dropout = drop_rate,
                                        beta = True,
                                        edge_dim = edge_dim,
                                        )
        self.convs[-1] = TransformerConv(num_hidden, num_hidden//num_heads,
                                        heads = num_heads,
                                        concat = False, 
                                        dropout = drop_rate,
                                        beta = True,
                                        edge_dim = edge_dim,
                                        )

    def forward(self, **inputs):
        batch_x = inputs.get("node_embedding")
        B, N = batch_x.shape[0], batch_x.shape[1]
        edge_indices = [edges + b*N for (b, edges) in zip(range(B), inputs.get("edge_indices"))]
        edge_index = torch.cat(edge_indices, 1)
        if self.edge_dim is not None:
            edge_attr = torch.cat(inputs.get("edge_attrs"))
        else:
            edge_attr = None
        x = batch_x.reshape(B*N, -1)
        
        for i in range(self.num_layers):
            x = self.convs[i](x, edge_index, edge_attr=edge_attr)
            if i != self.num_layers - 1:
                x = F.relu(self.norms[i](x))
        return x.reshape(B, N, -1)


class PNA(nn.Module):
    def __init__(self, num_layers, input_dim, hidden_dim, 
                 dropout, deg, edge_dim=None):
        super().__init__()

        aggregators = ['mean', 'min', 'max', 'std']
        scalers = ['identity', 'amplification', 'attenuation']

        self.num_layers = num_layers
        self.dropout = dropout
        self.edge_dim = edge_dim

        self.convs = nn.ModuleList()
        self.batch_norms = nn.ModuleList()
        self.convs.append(PNAConv(input_dim, hidden_dim,
                           aggregators=aggregators, scalers=scalers, deg=deg,
                           edge_dim=edge_dim, towers=4, pre_layers=1, post_layers=1,
                           divide_input=False))
        self.batch_norms.append(BatchNorm(hidden_dim))
        
        for _ in range(num_layers - 2):
            conv = PNAConv(hidden_dim, hidden_dim,
                           aggregators=aggregators, scalers=scalers, deg=deg,
                           edge_dim=edge_dim, towers=4, pre_layers=1, post_layers=1,
                           divide_input=False)
            self.convs.append(conv)
            self.batch_norms.append(BatchNorm(hidden_dim))

        self.convs.append(PNAConv(hidden_dim, hidden_dim,
                           aggregators=aggregators, scalers=scalers, deg=deg,
                           edge_dim=edge_dim, towers=4, pre_layers=1, post_layers=1,
                           divide_input=False))
        
        self.mlp = nn.Linear(hidden_dim, hidden_dim)

        self.reset_parameters()

    def reset_parameters(self):
        nn.init.xavier_normal_(self.mlp.weight)

    def forward(self, **inputs):
        batch_x = inputs.get("node_embedding")
        B, N = batch_x.shape[0], batch_x.shape[1]
        edge_indices = [edges + b*N for (b, edges) in zip(range(B), inputs.get("edge_indices"))]
        edge_index = torch.cat(edge_indices, 1)
        if self.edge_dim is not None:
            edge_attr = torch.cat(inputs.get("edge_attrs"))
        else:
            edge_attr = None
        x = batch_x.reshape(B*N, -1)
        
        for i, (conv, batch_norm) in enumerate(zip(self.convs, self.batch_norms)):
            x = conv(x, edge_index, edge_attr)
            if i != self.num_layers - 1:
                x = F.relu(batch_norm(x))
                x = F.dropout(x, p=self.dropout, training=self.training)
        x = self.mlp(x)
        return x.reshape(B, N, -1)


class GINEConv(MessagePassing):
    def __init__(self, num_layers, node_dim, edge_dim, dropout, train_eps=True):
        super(GINEConv, self).__init__(aggr='add')
        self.edge_dim = edge_dim
        if edge_dim is not None:
            self.edge_mlp = nn.Linear(edge_dim, node_dim)
        
        self.node_mlp = OrderedDict()
        for layer in range(num_layers):
            if layer == num_layers - 1:
                self.node_mlp[f'linear{layer}'] = nn.Linear(node_dim, node_dim)
            else:
                self.node_mlp[f'linear{layer}'] = nn.Linear(node_dim, node_dim)
                self.node_mlp[f'batchnorm{layer}'] = nn.BatchNorm1d(node_dim)
                self.node_mlp[f'relu{layer}'] = nn.ReLU()
                self.node_mlp[f'dropout{layer}'] = nn.Dropout(dropout)
        self.node_mlp = nn.Sequential(self.node_mlp)

        self.initial_eps = 0.
        if train_eps:
            self.eps = torch.nn.Parameter(torch.Tensor([self.initial_eps]))
        else:
            self.register_buffer('eps', torch.Tensor([self.initial_eps]))
        self.reset_parameters()
    
    def reset_parameters(self):
        if self.edge_dim is not None:
            nn.init.xavier_uniform_(self.edge_mlp.weight)
        reset(self.node_mlp)
        self.eps.data.fill_(self.initial_eps)
    
    def forward(self, x, edge_index, edge_attr, size = None)-> Tensor:
        if isinstance(x, Tensor):
            x: OptPairTensor = (x, x)
        
        if edge_attr is not None:
            edge_attr = self.edge_mlp(edge_attr)
        # propagate_type: (x: OptPairTensor, edge_attr: OptTensor)
        out = self.propagate(edge_index, x=x, edge_attr=edge_attr, size=size)

        x_r = x[1]
        if x_r is not None:
            out = out + (1 + self.eps) * x_r
        return self.node_mlp(out)
    
    def message(self, x_j: Tensor, edge_attr: OptTensor) -> Tensor:
        if edge_attr is None:
            return x_j.relu()
        else:
            return (x_j + edge_attr).relu()


class GINE(nn.Module):
    def __init__(self, num_layers, in_channels, hidden_channels, dropout, edge_dim=None):
        super().__init__()
        self.num_layers = num_layers
        self.dropout = dropout
        self.edge_dim = edge_dim

        self.conv = GINEConv(num_layers, node_dim=in_channels, edge_dim=edge_dim, 
                            dropout=dropout, train_eps=True)
    
    def forward(self, **inputs):
        batch_x = inputs.get("node_embedding")
        B, N = batch_x.shape[0], batch_x.shape[1]
        edge_indices = [edges + b*N for (b, edges) in zip(range(B), inputs.get("edge_indices"))]
        edge_index = torch.cat(edge_indices, 1)
        if self.edge_dim is not None:
            edge_attr = torch.cat(inputs.get("edge_attrs"))
        else:
            edge_attr = None
        x = batch_x.reshape(B*N, -1)
        
        x = self.conv(x, edge_index, edge_attr=edge_attr)
        return x.reshape(B, N, -1)