import torch
import torch.nn as nn
import torch.nn.functional as F
from typing import Dict, Optional

from torchmetrics import AUROC
from .gnn import GNNDecoder
from .utils import make_cdist_mask, mask_hidden_state, align_conformer_hat_to_conformer, make_mask_for_pyd_batch_graph
from torch_geometric.typing import Adj, OptTensor, SparseTensor
import itertools




def get_soft_label(distance, threshold=5.0, width=0.5):
    return 1 / (1 + torch.exp((distance - threshold) / width))


class ConformerPredictionHead(nn.Module):
    def __init__(self, hidden_X_dim: int = 256, width: float = 0.8) -> None:
        super().__init__()
        self.head = nn.Sequential(nn.Linear(hidden_X_dim, hidden_X_dim * 3), nn.GELU(), nn.Linear(hidden_X_dim * 3, 3))
        self.lin = nn.Linear(hidden_X_dim, 1)
        self.width = width

    def forward(self, conformer: torch.Tensor, hidden_X: torch.Tensor, padding_mask: torch.Tensor = None,
                compute_loss: bool = True, conformer_base: torch.Tensor = None,
                adj_t: torch.Tensor = None, near_edge_ratio: torch.Tensor = None) -> torch.Tensor:
        if conformer_base is not None:
            hidden_X = torch.stack((hidden_X, conformer_base), -1)
            attn_score = self.lin(hidden_X.transpose(3, 2)).squeeze(-1)  # (B, N, 2)
            scale = hidden_X.shape[2] ** 0.5
            attn_score = attn_score / torch.tensor(scale)
            attn_score = torch.softmax(attn_score, dim=-1).unsqueeze(2)
            hidden_X = torch.sum(attn_score * hidden_X, dim=-1)

        conformer_hat = self.head(hidden_X)
        conformer_hat = mask_hidden_state(conformer_hat, padding_mask)

        conformer_hat = align_conformer_hat_to_conformer(conformer_hat, conformer)
        conformer_hat = mask_hidden_state(conformer_hat, padding_mask)

        if not compute_loss:
            return conformer_hat

        result_dict = self._compute_loss(conformer, conformer_hat, padding_mask)

        # if adj_t is not None:
        #     B, N = conformer_hat.shape[0], conformer_hat.shape[1]  # shape of the conformer_hat: (B, N, 3)
        #     deg = near_edge_ratio.reshape(B, N, 1, 1)
        #     directions = conformer_hat.unsqueeze(2) - conformer_hat.unsqueeze(1)  # (B, N, N, 3)
        #     directions = F.normalize(directions, dim=-1)
        #     cos_sim = torch.einsum('bijc,bikc->bijk', directions, directions)  # (B, N, N, N)
        #     cos_sim = cos_sim * deg
        #
        #     adj = adj_t.clone().to(torch.bool)
        #     adj[:, torch.arange(N), torch.arange(N)] = 0
        #     neighbor_mask = adj.unsqueeze(3) & adj.unsqueeze(2)
        #     neighbor_mask[:, :, torch.arange(N), torch.arange(N)] = False
        #     cos_sim = cos_sim[neighbor_mask]
        #     loss_cos = F.relu(cos_sim).mean()
        #     result_dict["loss"] += loss_cos

        # if adj_t is not None:
        #     B, N = conformer_hat.shape[0], conformer_hat.shape[1]  # Shape: (B, N, 3)
        #     epsilon = 1e-6  # Small constant to avoid division by zero
        #
        #     # Step 1: Compute bond vectors and lengths
        #     directions = conformer_hat.unsqueeze(2) - conformer_hat.unsqueeze(1)  # (B, N, N, 3)
        #     bond_lengths = torch.norm(directions, dim=-1)  # (B, N, N)
        #
        #     # Zero out self-distances without in-place operations
        #     mask_self = torch.eye(N, device=conformer_hat.device, dtype=torch.bool).unsqueeze(0)  # Shape: (1, N, N)
        #     mask_self = mask_self.expand(B, -1, -1)  # Expand to match batch size
        #     bond_lengths = bond_lengths.masked_fill(mask_self, 0)
        #
        #     # Step 2: Compute inverse distances and apply adjacency mask
        #     inverse_distances = 1 / (bond_lengths + epsilon)  # (B, N, N)
        #     adj = adj_t.clone().to(torch.bool)
        #
        #     # Remove self-loops without in-place operation
        #     adj = adj.masked_fill(mask_self, False)
        #     inverse_distances = inverse_distances * adj  # Consider only bonded atoms
        #
        #     # Step 3: Perform atom-wise normalization
        #     sum_inverse_distances = inverse_distances.sum(dim=-1, keepdim=True) + epsilon  # (B, N, 1)
        #     normalized_inverse_distances = inverse_distances / sum_inverse_distances  # (B, N, N)
        #
        #     # Step 4: Compute weights for pairs of bonds
        #     weights = normalized_inverse_distances.unsqueeze(3) * normalized_inverse_distances.unsqueeze(
        #         2)  # (B, N, N, N)
        #
        #     # Step 5: Normalize direction vectors
        #     directions_normalized = F.normalize(directions, dim=-1)  # (B, N, N, 3)
        #
        #     # Step 6: Compute cosine similarities
        #     cos_sim = torch.einsum('bijc,bikc->bijk', directions_normalized, directions_normalized)  # (B, N, N, N)
        #
        #     # Step 7: Prepare neighbor mask for bond pairs
        #     neighbor_mask = adj.unsqueeze(3) & adj.unsqueeze(2)  # (B, N, N, N)
        #
        #     # Exclude pairs where j == k without in-place operation
        #     mask_jk = torch.eye(N, device=conformer_hat.device, dtype=torch.bool).unsqueeze(0).unsqueeze(
        #         0)  # (1, 1, N, N)
        #     mask_jk = mask_jk.expand(B, N, -1, -1)
        #     neighbor_mask = neighbor_mask & (~mask_jk)
        #
        #     # Step 8: Extract relevant cosine similarities and weights
        #     cos_sim_values = cos_sim[neighbor_mask]  # (num_entries,)
        #     weights_values = weights[neighbor_mask]  # (num_entries,)
        #
        #     # Step 9: Compute the weighted loss term
        #     loss_cos = (F.relu(cos_sim_values) * weights_values).mean()
        #
        #     # Step 10: Update the total loss
        #     result_dict["loss"] += loss_cos
        return result_dict

    def _compute_loss(self, conformer: torch.Tensor, conformer_hat: torch.Tensor,
                      padding_mask: torch.Tensor = None) -> torch.Tensor:
        # Convenient for design loss function in the future.
        cdist, cdist_hat = torch.cdist(conformer, conformer), torch.cdist(conformer_hat, conformer_hat)
        c_dist_mask = make_cdist_mask(padding_mask)
        cdist, cdist_hat = cdist * c_dist_mask, cdist_hat * c_dist_mask
        cdist_mae, loss = self._compute_cdist_mae(cdist, cdist_hat, c_dist_mask)
        cdist_mse = self._compute_cdist_mse(cdist, cdist_hat, c_dist_mask)
        coord_rmsd = self._compute_conformer_rmsd(conformer, conformer_hat, padding_mask)
        # compute learnable weighted loss
        # loss = self.criterion([cdist_mae, cdist_rmse])
        return {
            "loss": loss,
            "cdist_mae": cdist_mae.detach(),
            "cdist_mse": cdist_mse.detach(),
            "coord_rmsd": coord_rmsd.detach(),
            "conformer": conformer.detach(),
            "conformer_hat": conformer_hat.detach(),
        }

    @staticmethod
    def _compute_cdist_mae(masked_cdist: torch.Tensor, masked_cdist_hat: torch.Tensor,
                           cdist_mask: torch.Tensor) -> torch.Tensor:
        """Compute mean absolute error of conformer and conformer_hat.

        Args:
            - masked_cdist (torch.Tensor): A torch tensor of shape (b, l, l), which denotes the pairwise distance matrix of the conformer.
            - masked_cdist_hat (torch.Tensor): A torch tensor of shape (b, l, l), which denotes the pairwise distance matrix of the conformer_hat.
            - cdist_mask (torch.Tensor): A torch tensor of shape (b, l, l), which denotes the mask of the pairwise distance matrix.

        Returns:
            torch.Tensor: The mean absolute error of conformer and conformer_hat.
        """
        loss = F.l1_loss(masked_cdist, masked_cdist_hat, reduction="sum") / cdist_mask.sum()  # exclude padding atoms
        mae = loss
        return mae, loss

    @staticmethod
    def _compute_cdist_mse(masked_cdist: torch.Tensor, masked_cdist_hat: torch.Tensor,
                           cdist_mask: torch.Tensor) -> torch.Tensor:
        """Compute root mean squared error of conformer and conformer_hat.

        Args:
            - masked_cdist (torch.Tensor): A torch tensor of shape (b, l, l), which denotes the pairwise distance matrix of the conformer.
            - masked_cdist_hat (torch.Tensor): A torch tensor of shape (b, l, l), which denotes the pairwise distance matrix of the conformer_hat.
            - cdist_mask (torch.Tensor): A torch tensor of shape (b, l, l), which denotes the mask of the pairwise distance matrix.

        Returns:
            torch.Tensor: The root mean squared error of conformer and conformer_hat.
        """
        mse = F.mse_loss(masked_cdist, masked_cdist_hat, reduction="sum") / cdist_mask.sum()  # exclude padding atoms
        return mse

    @staticmethod
    def _compute_conformer_rmsd(masked_conformer: torch.Tensor, masked_conformer_hat: torch.Tensor,
                                padding_mask: torch.Tensor) -> torch.Tensor:
        """Compute root mean squared deviation of conformer and conformer_hat.

        Args:
            - masked_conformer (torch.Tensor): A torch tensor of shape (b, l, 3), which denotes the coordinate of the conformer.
            - masked_conformer_hat (torch.Tensor): A torch tensor of shape (b, l, 3), which denotes the coordinate of the conformer_hat.
            - padding_mask (torch.Tensor): A torch tensor of shape (b, l), which denotes the mask of the conformer.

        Returns:
            torch.Tensor: The root mean squared deviation of conformer and conformer_hat.
        """
        R, R_h, M = masked_conformer, masked_conformer_hat, padding_mask
        delta = (R - R_h).to(torch.float32)
        point_2_norm = torch.norm(delta, p=2, dim=-1)
        MSD = torch.sum(point_2_norm ** 2, dim=-1) / torch.sum(M, dim=-1)
        RMSD = torch.sqrt(MSD)
        return RMSD.mean()


class GNNConformerPredictionHead(nn.Module):
    def __init__(self, hidden_X_dim: int = 300) -> None:
        super().__init__()
        self.head = nn.Sequential(nn.Linear(hidden_X_dim, hidden_X_dim * 3), nn.ReLU(), nn.Linear(hidden_X_dim * 3, 3))
        self.lin = nn.Linear(hidden_X_dim, 1)

    def forward(self, conformer: torch.Tensor, hidden_X: torch.Tensor, batch: torch.Tensor,
                conformer_base: torch.Tensor = None):
        if conformer_base is not None:
            hidden_X = torch.stack((hidden_X, conformer_base), -1)  # N, d, 2
            attn_score = self.lin(hidden_X.transpose(2, 1)).squeeze(-1)  # (N, 2)
            scale = hidden_X.shape[2] ** 0.5
            attn_score = attn_score / torch.tensor(scale)
            attn_score = torch.softmax(attn_score, dim=-1).unsqueeze(1)
            # import pdb; pdb.set_trace()
            hidden_X = torch.sum(attn_score * hidden_X, dim=-1)

        conformer_hat = self.head(hidden_X)
        return self._compute_loss(conformer, conformer_hat, batch)

    def _compute_loss(self, conformer: torch.Tensor, conformer_hat: torch.Tensor, batch: torch.Tensor):
        cdist = torch.cdist(conformer, conformer)
        cdist_hat = torch.cdist(conformer_hat, conformer_hat)
        M = make_mask_for_pyd_batch_graph(batch)
        masked_cdist, masked_cdist_hat = cdist * M, cdist_hat * M  # only consider pairs in the same molecule
        mae = self._compute_cdist_mae(masked_cdist, masked_cdist_hat, M)
        mse = self._compute_cdist_mse(masked_cdist, masked_cdist_hat, M)
        num_samples = len(torch.unique(batch))
        coord_rmsd = self._compute_conformer_rmsd(conformer, conformer_hat, batch, num_samples)
        loss = mae

        return {
            "loss": loss,
            "cdist_mae": mae.detach(),
            "cdist_mse": mse.detach(),
            "coord_rmsd": coord_rmsd.detach(),
            "conformer": conformer.detach(),
            "conformer_hat": conformer_hat.detach(),
            "mask": M.detach(),
        }

    @staticmethod
    def _compute_cdist_mae(masked_cdist: torch.Tensor, masked_cdist_hat: torch.Tensor,
                           cdist_mask: torch.Tensor) -> torch.Tensor:
        D, D_h, M = masked_cdist, masked_cdist_hat, cdist_mask
        mae = F.l1_loss(D, D_h, reduction="sum") / M.sum()
        return mae

    @staticmethod
    def _compute_cdist_mse(masked_cdist: torch.Tensor, masked_cdist_hat: torch.Tensor,
                           cdist_mask: torch.Tensor) -> torch.Tensor:
        D, D_h, M = masked_cdist, masked_cdist_hat, cdist_mask
        mse = F.mse_loss(D, D_h, reduction="sum") / M.sum()
        return mse

    @staticmethod
    def _compute_conformer_rmsd(masked_conformer: torch.Tensor, masked_conformer_hat: torch.Tensor, batch: torch.Tensor,
                                num_samples: int) -> torch.Tensor:
        """Compute root mean squared deviation of conformer and conformer_hat.

        Args:
            - masked_conformer (torch.Tensor): A torch tensor of shape (num_atoms, 3), which denotes the coordinate of molecules' conformer.
            - masked_conformer_hat (torch.Tensor): A torch tensor of shape (num_atoms, 3), which denotes the coordinate of molecules' conformer_hat.

        Returns:
            torch.Tensor: The root mean squared deviation of conformer and conformer_hat.
        """
        R, R_h = masked_conformer, masked_conformer_hat
        delta = (R - R_h).to(torch.float32)
        point_2_norm = torch.norm(delta, p=2, dim=-1)
        RMSD = [torch.sqrt(torch.sum(point_2_norm[batch == i] ** 2) / (batch == i).sum()) for i in range(num_samples)]
        RMSD = torch.tensor(RMSD)
        # MSD = torch.sum(point_2_norm**2, dim=-1) / point_2_norm.shape[0]
        # RMSD = torch.sqrt(MSD)
        return RMSD.mean()


class GNNProximityPredictionHead(nn.Module):
    def __init__(self, hidden_X_dim: int = 300, JK: str = "last", num_layers: int = 8) -> None:
        super().__init__()
        self.JK = JK
        if self.JK == "concat":
            self.head = nn.Sequential(
                nn.Linear(hidden_X_dim * num_layers, hidden_X_dim * num_layers // 2),
                nn.ReLU(),
                nn.Linear(hidden_X_dim * num_layers // 2, hidden_X_dim)
            )

    def forward(self, edge_pos: Optional[torch.Tensor], edge_neg: Optional[torch.Tensor],
                hidden_X: torch.Tensor, margin: int, conformer: torch.Tensor,
                batch: torch.Tensor):

        diff = conformer.unsqueeze(1) - conformer.unsqueeze(0)
        dist_matrix = torch.sqrt((diff ** 2).sum(-1))

        if self.JK == "concat":
            hidden_X = self.head(hidden_X)

        if edge_pos is not None and edge_pos.numel() and edge_neg.numel():
            pos_out = torch.sum(hidden_X[edge_pos[0]] * hidden_X[edge_pos[1]], dim=-1).unsqueeze(-1)
            neg_out = torch.sum(hidden_X[edge_neg[0]] * hidden_X[edge_neg[1]], dim=-1).unsqueeze(-1)

            labels = torch.zeros(pos_out.shape[0] + neg_out.shape[0],
                                 dtype=torch.float, device=edge_pos.device).unsqueeze(-1)
            labels[:pos_out.shape[0]] = 1

            if margin > 0:
                pos_labels = get_soft_label(dist_matrix[edge_pos[0], edge_pos[1]], width=margin)
                neg_labels = get_soft_label(dist_matrix[edge_neg[0], edge_neg[1]], width=margin)
                soft_labels = torch.cat([pos_labels, neg_labels]).unsqueeze(-1)
                loss = F.binary_cross_entropy_with_logits(torch.cat([pos_out, neg_out], dim=0), soft_labels)
            else:
                loss = F.binary_cross_entropy_with_logits(torch.cat([pos_out, neg_out], dim=0), labels)

            accuracy = (torch.sigmoid(pos_out) > 0.5).sum() + (torch.sigmoid(neg_out) < 0.5).sum()
            accuracy = accuracy / (edge_pos.shape[1] + edge_neg.shape[1])
            false_negative = (torch.sigmoid(pos_out) < 0.5).sum() / edge_pos.shape[1]

            num_samples = edge_pos.shape[1] + edge_neg.shape[1]

        else:
            loss = torch.tensor(0.)
            accuracy = torch.tensor(0.)
            num_samples = 0
            false_negative = torch.tensor(0.)
            pos_out = torch.tensor(0.)
            neg_out = torch.tensor(0.)

        return {
            "pred": hidden_X.detach(),
            "loss": loss,
            "acc": accuracy.detach(),
            "num_samples": num_samples,
            "false_negative": false_negative.detach(),
            "min_pos_logit": pos_out.min().item(),
            "max_neg_logit": neg_out.max().item()
        }


class GraphReConstructionHead(nn.Module):
    num_node_type = 119
    num_node_chiral = 4
    num_edge_type = 4

    def __init__(self, in_dim: int = 300, hidden_dim: int = 600, dropout: float = 0.0,
                 re_build_edge: bool = False) -> None:
        super().__init__()
        self.re_build_edge = re_build_edge
        self.atom_re_constructor = GNNDecoder(
            in_dim=in_dim,
            hidden_dim=hidden_dim,
            out_dim=self.num_node_type,
            dropout=dropout,
        )
        self.atom_chiral_re_constructor = GNNDecoder(
            in_dim=in_dim,
            hidden_dim=hidden_dim,
            out_dim=self.num_node_chiral,
            dropout=dropout,
        )
        if re_build_edge:
            self.edge_re_constructor = GNNDecoder(
                in_dim=in_dim,
                hidden_dim=hidden_dim,
                out_dim=self.num_edge_type,
                dropout=dropout,
            )

    def forward(self, **inputs) -> Dict[str, torch.Tensor]:
        node_representation = inputs.get("node_representation")
        node_type = inputs.get("node_type")
        node_chiral_type = inputs.get("node_chiral_type")
        edge_type = inputs.get("edge_type")
        edge_dire_type = inputs.get("edge_dire_type")
        edge_index = inputs.get("edge_index")

        node_logits = self.atom_re_constructor(
            node_attr=node_representation,
            edge_index=edge_index,
            edge_type=edge_type,
            edge_dire_type=edge_dire_type,
        )
        node_chiral_logits = self.atom_chiral_re_constructor(
            node_attr=node_representation, edge_index=edge_index, edge_type=edge_type, edge_dire_type=edge_dire_type
        )
        edge_logits = None
        if self.re_build_edge:
            edge_representation = node_representation[edge_index[0]] + node_representation[edge_index[1]]
            edge_logits = self.edge_re_constructor(node_attr=edge_representation, edge_index=edge_index,
                                                   edge_type=edge_type, edge_dire_type=edge_dire_type)
        reconstruction_loss = self._compute_loss(
            node_logits,
            node_chiral_logits,
            edge_logits,
            node_type,
            node_chiral_type,
            edge_type,
        )
        reconstruction_accuracy = self._compute_accuracy(node_logits, node_chiral_logits, edge_logits, node_type,
                                                         node_chiral_type, edge_type)
        return {"reconstruction_loss": reconstruction_loss, "reconstruction_accuracy": reconstruction_accuracy.detach()}

    def _compute_loss(self, node_logits, node_chiral_logits, edge_logits, node_type, node_chiral_type, edge_type):
        node_type_loss = F.cross_entropy(node_logits, node_type)
        node_chiral_loss = F.cross_entropy(node_chiral_logits, node_chiral_type)
        if self.re_build_edge:
            edge_type_loss = F.cross_entropy(edge_logits, edge_type)
            return node_type_loss + node_chiral_loss + edge_type_loss
        return node_type_loss + node_chiral_loss

    def _compute_accuracy(self, node_logits, node_chiral_logits, edge_logits, node_type, node_chiral_type, edge_type):
        node_type_preds, node_chiral_preds = node_logits.detach().argmax(dim=-1), node_chiral_logits.detach().argmax(
            dim=-1)
        node_type_accuracy = (node_type_preds == node_type.detach()).float().mean()
        node_chiral_accuracy = (node_chiral_preds == node_chiral_type.detach()).float().mean()
        if self.re_build_edge:
            edge_type_preds = edge_logits.detach().argmax(dim=-1)
            edge_type_accuracy = (edge_type_preds == edge_type.detach()).float().mean()
            # return (node_type_accuracy + node_chiral_accuracy + edge_type_accuracy) / 3
        # return (node_type_accuracy + node_chiral_accuracy) / 2
        return node_type_accuracy  # cause the node_type_accuracy and node_chiral_accuracy grow fast, so we only use node_type_accuracy


class GraphRegressionHead(nn.Module):
    def __init__(self, hidden_X_dim: int = 256) -> None:
        super().__init__()
        self.head = nn.Sequential(nn.Linear(hidden_X_dim, hidden_X_dim * 3), nn.GELU(), nn.Linear(hidden_X_dim * 3, 1))

    def forward(self, hidden_X: torch.tensor, labels: torch.Tensor) -> torch.Tensor:
        logits = self.head(hidden_X).squeeze(-1)
        loss = F.l1_loss(logits, labels)
        mse = F.mse_loss(logits, labels)
        return {
            "loss": loss,
            "mae": loss.detach(),
            "mse": mse.detach(),
            "logits": logits.detach(),
            "labels": labels.detach(),
        }
