"""
    Collator for Conformer.
"""

import torch
import torch.nn.functional as F
import numpy as np
from typing import Dict, List, Sequence
from ..modules.utils_pe_laplacian  import AddLaplacianEigenvectorPE, AddRandomWalkPE
from torch_geometric.data import Data
from rdkit import Chem
from ..modules.utils import get_sigma_and_epsilon
from ..modules.collating_utils import get_soft_label, print_bond_distances, get_adjacency, valid_length_to_mask, get_laplacian_eigenvectors
import wandb
from copy import deepcopy
from torch_geometric.utils import degree, to_undirected
from sklearn.cluster import KMeans
import warnings

warnings.simplefilter("ignore", UserWarning)


atomic_symbol_to_number = {
    'H': 0, 'C': 5, 'N': 6, 'O': 7, 'F': 8
}

atomic_number_to_symbol = {
    0: 'H', 5: 'C', 6: 'N', 7: 'O', 8: 'F'
}

bond_type_map = {
    Chem.BondType.SINGLE: 0,
    Chem.BondType.DOUBLE: 1,
    Chem.BondType.TRIPLE: 2,
    Chem.BondType.AROMATIC: 3
}

class Collator:
    def __init__(self, max_nodes: int = None, max_edges: int = None, laplacian_k: int = 5, 
                 edge_type: str = 'default', prox_collator=None, prox_predictor = None) -> None:
        self.max_nodes = max_nodes
        self.max_edges = max_edges
        self.edge_type = edge_type
        self.keys = None
        self.max_degree = None
        self.laplacian_k = laplacian_k

        self.pe_f = AddLaplacianEigenvectorPE(
            k=laplacian_k, 
            attr_name="pe", 
            is_undirected=True, 
            normalize="sym", 
        )

    @torch.no_grad()
    def __call__(self, mol_sq: Sequence[Dict]) -> Dict[str, torch.Tensor]:
        keys = mol_sq[0].keys()
        self.keys = keys
        mol_sq = [self._transform(self._get_pyg_data(mol)) for mol in mol_sq]

        num_mol = len(mol_sq)
        num_nodes = [len(mol.node_type) for mol in mol_sq]
        max_nodes = max(num_nodes) if self.max_nodes is None else self.max_nodes
        node_type = torch.zeros((num_mol, max_nodes), dtype=torch.long)
        
        lap_eigenvectors = torch.zeros((num_mol, max_nodes, max_nodes), dtype=torch.float32)
        adjacency = torch.zeros((num_mol, max_nodes, max_nodes), dtype=torch.long)
        near_adjacency = torch.zeros((num_mol, max_nodes, max_nodes), dtype=torch.long)
        adjacency_t0 = torch.zeros((num_mol, max_nodes, max_nodes), dtype=torch.long)
        adjacency_t1 = torch.zeros((num_mol, max_nodes, max_nodes), dtype=torch.long)
        adjacency_t2 = torch.zeros((num_mol, max_nodes, max_nodes), dtype=torch.long)
        adjacency_t3 = torch.zeros((num_mol, max_nodes, max_nodes), dtype=torch.long)

        sigma = torch.zeros((num_mol, max_nodes, max_nodes), dtype=torch.float32)
        epsilon = torch.zeros((num_mol, max_nodes, max_nodes), dtype=torch.float32)
        
        node_input_ids = torch.zeros((num_mol, max_nodes), dtype=torch.long) if "input_ids" in keys else None
        conformer = torch.zeros((num_mol, max_nodes, 3), dtype=torch.float)  # if "conformer" in keys else None
        labels = torch.tensor([mol.labels for mol in mol_sq], dtype=torch.float) if "labels" in keys else None
        node_attr = torch.zeros((num_mol, max_nodes, 9), dtype=torch.long) if "node_attr" in keys else None
        pred_dist = torch.zeros((num_mol, max_nodes, max_nodes), dtype=torch.float32)
        near_edge_ratio = torch.zeros((num_mol, max_nodes), dtype=torch.long)

        edge_attr = torch.zeros((num_mol, max_nodes, max_nodes, 4), dtype=torch.long)
        num_near_edges = torch.zeros((num_mol, max_nodes), dtype=torch.long)

        for i, mol in enumerate(mol_sq):

            if self.edge_type == "multi_relational_degree":
                pred_dist[i, :num_nodes[i], :num_nodes[i]] = mol.pred_dist
            
            adj = get_adjacency(num_nodes[i], mol.edge_index)
            adjacency[i, :num_nodes[i], :num_nodes[i]] = adj

            mol = self.pe_f(mol)
            pe_dim = mol.pe.shape[1]
            lap_eigenvectors[i, :num_nodes[i], :min(max_nodes, pe_dim)] = mol.pe[:, :min(max_nodes, pe_dim)]
        
            if hasattr(mol, "sigma") and hasattr(mol, "epsilon"):
                sigma[i, :num_nodes[i], :num_nodes[i]] = mol.sigma
                epsilon[i, :num_nodes[i], :num_nodes[i]] = mol.epsilon

            if mol.edge_index.numel():
                row, col = mol.edge_index
                edge_attr[i, row, col] = mol.edge_attr
                num_near_edges[i, :num_nodes[i]] = mol.num_near_edges
                    
                for typ in [0, 1, 2, 3]: # bond
                    tmp_edges = mol.edge_index[:, mol.edge_type==typ]
                    row, col = tmp_edges
                    if typ == 0:
                        adjacency_t0[i, row, col] = 1
                    elif typ == 1:
                        adjacency_t1[i, row, col] = 1
                    elif typ == 2:
                        adjacency_t2[i, row, col] = 1
                    elif typ == 3:
                        adjacency_t3[i, row, col] = 1
        
            node_type[i, : num_nodes[i]] = mol.node_type + 1  # 0 for padding
            
            if "input_ids" in keys:
                node_input_ids[i, : num_nodes[i]] = mol.input_ids + 1  # 0 for padding
            if "conformer" in keys:
                conformer[i, : num_nodes[i]] = mol.conformer
            if "node_attr" in keys:
                node_attr[i, : num_nodes[i]] = mol.node_attr + 1  # 0 for padding

        res_dic = {
            "node_type": node_type,
            "node_mask": valid_length_to_mask(num_nodes, max_nodes),
            "adjacency": adjacency,
            "near_adjacency": near_adjacency,
            "lap_eigenvectors": lap_eigenvectors,
            "adjacency_t0": adjacency_t0,
            "adjacency_t1": adjacency_t1,
            "adjacency_t2": adjacency_t2,
            "adjacency_t3": adjacency_t3,
            "pred_dist": pred_dist,
            "edge_attr": edge_attr,
            "num_near_edges": num_near_edges,
            "near_edge_ratio": near_edge_ratio,
            "sigma": sigma,
            "epsilon": epsilon,
        }

        if node_input_ids is not None:
            res_dic["node_input_ids"] = node_input_ids
        if labels is not None:
            res_dic["labels"] = labels
        if conformer is not None:
            res_dic["conformer"] = conformer
        if node_attr is not None:
            res_dic["node_attr"] = node_attr
        
        return res_dic

    def _get_pyg_data(self, mol: Dict) -> Data:
        return Data(
            node_type=torch.tensor(mol["node_type"], dtype=torch.long),
            edge_index=torch.tensor(mol["edge_index"], dtype=torch.long),
            node_attr=torch.tensor(mol["node_attr"], dtype=torch.long) if "node_attr" in self.keys else None,
            edge_attr=torch.tensor(mol["edge_attr"], dtype=torch.long),
            conformer=torch.tensor(mol["conformer"], dtype=torch.float32),
            labels=torch.tensor(mol["labels"], dtype=torch.float32) if "labels" in self.keys else None,
            input_ids=torch.tensor(mol["input_ids"], dtype=torch.long) if "input_ids" in self.keys else None,
            pred = torch.tensor(mol["representation"], dtype=torch.float32) if "representation" in mol.keys() else None,
        )

    def _transform(self, data: Data) -> Data:
        if not data.edge_attr.numel():
            data.edge_type = torch.empty((0), dtype=torch.long)
        elif len(data.edge_attr.shape) == 1:
            data.edge_type = data.edge_attr[0]
        else:
            data.edge_type = data.edge_attr[:, 0]

        data.edge_index_merged = data.edge_index
        d = degree(data.edge_index[1], num_nodes=len(data.node_type), dtype=torch.long)
        max_degree = d.max().item()
        num_near_edges = max_degree - d
        data.num_near_edges = num_near_edges
        data.num_near_edges = num_near_edges
        data.near_edge_ratio = num_near_edges / max_degree

        epsilon, sigma = get_sigma_and_epsilon(data)
        data.epsilon = torch.sqrt((epsilon.unsqueeze(1) * epsilon.unsqueeze(0)))
        data.sigma = ((sigma.unsqueeze(1) + sigma.unsqueeze(0))/2)

        return data