"""
    This module contains self-defined GraphBert model.
"""

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch_geometric.nn as gnn
from .configuration_ours import Config

from ..modules import ConformerPredictionHead, GraphRegressionHead
from ..modules import ConformerPredictionOutput, GraphRegressionOutput
from ..modules import MultiHeadAttention, AddNorm, PositionWiseFFN, Residual
from ..modules import UniMP, PNA, GINE
from ..modules.utils import make_cdist_mask, compute_distance_residual_bias
from ..modules import AtomEmbedding, NodeEmbedding, BondEmbedding  # for Ogb embedding ablation
from ..modules.collating_utils import get_soft_label, retain_top_k
from ..modules.utils_pe_laplacian  import AddLaplacianEigenvectorPE
from transformers import PretrainedConfig, PreTrainedModel
from torch_geometric.data import Data


class ResidualAttentionBlock(nn.Module):
    def __init__(self, config: PretrainedConfig = None, encoder: bool = True) -> None:
        super().__init__()
        self.config = config
        self.encoder = encoder
        self.edge_type = getattr(config, "edge_type", "default")
        assert config is not None, f"config must be specified to build {self.__class__.__name__}"
        self.use_A_in_attn = getattr(config, "encoder_use_A_in_attn", False) if encoder else getattr(config, "decoder_use_A_in_attn", False)
        self.use_D_in_attn = getattr(config, "encoder_use_D_in_attn", False) if encoder else getattr(config, "decoder_use_D_in_attn", False)
        self.use_node_in_attn = getattr(config, "encoder_use_node_embed", False) if encoder else getattr(config, "decoder_use_node_embed", False)

        if encoder:
            num_edge_type = 1
        else:
            num_edge_type = 2

        self.multi_attention = MultiHeadAttention(
            d_q=getattr(config, "d_q", 256),
            d_k=getattr(config, "d_k", 256),
            d_v=getattr(config, "d_v", 256),
            d_model=getattr(config, "d_model", 256),
            n_head=getattr(config, "n_head", 8),
            qkv_bias=getattr(config, "qkv_bias", True),
            attn_drop=getattr(config, "attn_drop", 0.1),
            use_adjacency=self.use_A_in_attn,
            use_distance=self.use_D_in_attn,
            use_decomp=True,
            num_edge_type=num_edge_type,
        )
        self.add_norm01 = AddNorm(norm_shape=getattr(config, "d_model", 256), dropout=getattr(config, "norm_drop", 0.1),
                                  pre_ln=getattr(config, "pre_ln", True))
        self.position_wise_ffn = PositionWiseFFN(
            d_in=getattr(config, "d_model", 256),
            d_hidden=getattr(config, "d_ffn", 1024),
            d_out=getattr(config, "d_model", 256),
            dropout=getattr(config, "ffn_drop", 0.1),
        )
        self.add_norm02 = AddNorm(norm_shape=getattr(config, "d_model", 256), dropout=getattr(config, "norm_drop", 0.1),
                                  pre_ln=getattr(config, "pre_ln", True))

    def forward(self, **inputs):
        X, M = inputs.get("node_embedding"), inputs.get("node_mask")
        if self.use_A_in_attn:
            A = inputs.get("adjacency")
        else:
            if self.encoder:
                A = [
                    inputs.get("adjacency").type(torch.float32),
                ]
            else:
                A = [
                    inputs.get("adjacency").type(torch.float32),
                    inputs.get("force_adjacency").type(torch.float32),
                ]
        D = inputs.get("distance") if self.use_D_in_attn else None
        attn_out = self.multi_attention(X, X, X, attention_mask=M, adjacency_matrix=A, distance_matrix=D)
        Y, attn_weight = attn_out["out"], attn_out["attn_weight"]
        X = self.add_norm01(X, Y)
        Y = self.position_wise_ffn(X)
        X = self.add_norm02(X, Y)
        return {
            "out": X,
            "attn_weight": attn_weight,
        }


class OursPretrainedModel(PreTrainedModel):
    config_class = Config
    base_model_prefix = "Conformer"
    is_parallelizable = False
    main_input_name = "node_attr"

    def __init_weights__(self):
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.xavier_normal_(m.weight)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.Embedding):
                nn.init.xavier_normal_(m.weight)


class OursEncoder(OursPretrainedModel):
    def __init__(self, config: PretrainedConfig = None, *inputs, **kwargs):
        super().__init__(config, *inputs, **kwargs)
        assert config is not None, f"config must be specified to build {self.__class__.__name__}"
        self.embed_style = getattr(config, "embed_style", "atom_tokenized_ids")
        self.edge_type = getattr(config, "edge_type", "default")

        if self.embed_style == "atom_tokenized_ids":
            self.node_embedding = nn.Embedding(getattr(config, "atom_vocab_size", 513), getattr(config, "d_embed", 256), padding_idx=0)
        elif self.embed_style == "atom_type_ids":
            self.node_embedding = AtomEmbedding(atom_embedding_dim=getattr(config, "d_embed", 256), attr_reduction="sum")
        elif self.embed_style == "ogb":
            self.ogb_node_embedding = NodeEmbedding(atom_embedding_dim=getattr(config, "d_embed", 256), attr_reduction="sum")  # for Ogb embedding ablation

        self.encoder_blocks = nn.ModuleList([ResidualAttentionBlock(config, encoder=True) for _ in range(getattr(config, "n_encode_layers", 6))])
        self.__init_weights__()

    def forward(self, **inputs):
        # Using kwargs to make getting inputs more flexible
        if self.embed_style == "atom_tokenized_ids":
            node_input_ids = inputs.get("node_input_ids")
            node_embedding = self.node_embedding(node_input_ids)
        elif self.embed_style == "atom_type_ids":
            node_input_ids = inputs.get("node_attr")
            node_embedding = self.node_embedding(node_input_ids)
        elif self.embed_style == "ogb":
            node_embedding = self.ogb_node_embedding(inputs["node_attr"])  # for Ogb embedding ablation
        else:
            raise ValueError(f"Unknown embedding style: {self.embed_style}")
        # laplacian positional embedding
        lap = inputs.get("lap_eigenvectors")
        dim = node_embedding.shape[-1]
        node_embedding[:, :, : min(dim, lap.shape[-1])] = node_embedding[:, :, : min(dim, lap.shape[-1])] + lap[:, :, : min(dim, lap.shape[-1])]
        inputs["node_embedding"] = node_embedding

        if self.config.encoder_use_D_in_attn:
            C = inputs.get("conformer")
            D, D_M = torch.cdist(C, C), make_cdist_mask(inputs.get("node_mask"))
            D = compute_distance_residual_bias(cdist=D, cdist_mask=D_M)  # masked and row-max subtracted distance matrix
            inputs["distance"] = D

        attn_weight_dict = {}
        for i, encoder_block in enumerate(self.encoder_blocks):
            block_out = encoder_block(**inputs)
            node_embedding, attn_weight = block_out["out"], block_out["attn_weight"]
            inputs["node_embedding"] = node_embedding
            attn_weight_dict[f"encoder_block_{i}"] = attn_weight

        return {"node_embedding": node_embedding, "attn_weight_dict": attn_weight_dict}


class OursDecoder(OursPretrainedModel):
    def __init__(self, config: PretrainedConfig = None, *inputs, **kwargs):
        super().__init__(config, *inputs, **kwargs)
        assert config is not None, f"config must be specified to build {self.__class__.__name__}"
        self.decoder_blocks = nn.ModuleList([ResidualAttentionBlock(config, encoder=False) for _ in range(getattr(config, "n_decode_layers", 6))])
        self.__init_weights__()

    def forward(self, **inputs):
        node_embedding, lap = inputs.get("node_embedding"), inputs.get("lap_eigenvectors")
        # laplacian positional embedding
        dim = node_embedding.shape[-1]
        node_embedding[:, :, : min(dim, lap.shape[-1])] = node_embedding[:, :, : min(dim, lap.shape[-1])] + lap[:, :, : min(dim, lap.shape[-1])]
        inputs["node_embedding"] = node_embedding

        attn_weight_dict = {}
        for i, decoder_block in enumerate(self.decoder_blocks):
            block_out = decoder_block(**inputs)
            node_embedding, attn_weight = block_out["out"], block_out["attn_weight"]
            inputs["node_embedding"] = node_embedding
            attn_weight_dict[f"decoder_block_{i}"] = attn_weight
        return {"node_embedding": node_embedding, "attn_weight_dict": attn_weight_dict}

class OursConformerPrediction(OursPretrainedModel):
    def __init__(self, config: PretrainedConfig, *inputs, **kwargs):
        super().__init__(config, *inputs, **kwargs)
        self.edge_type = getattr(config, "edge_type", "default")
        self.width = getattr(config, "width", 0.8)

        self.encoder = OursEncoder(config)
        self.decoder = OursDecoder(config)
        self.conformer_head = ConformerPredictionHead(hidden_X_dim=getattr(config, "d_model", 256),
                                                      width=self.width)
        self.residual_head = ConformerPredictionHead(hidden_X_dim=getattr(config, "d_model", 256),
                                                     width=self.width)

        self.__init_weights__()

    def forward(self, **inputs):
        conformer, node_mask = inputs.get("conformer"), inputs.get("node_mask")

        encoder_out = self.encoder(**inputs)
        node_embedding = encoder_out["node_embedding"] #, encoder_out["attn_weight_dict"]

        cache_out = self.conformer_head(conformer=conformer, hidden_X=node_embedding, padding_mask=node_mask,
                                        compute_loss=True,
                                        near_edge_ratio=inputs.get(
                                            "near_edge_ratio"))  # , adj_t=inputs["adjacency"]

        loss_cache, conformer_cache = cache_out["loss"], cache_out["conformer_hat"]
        D_cache = inputs["pred_dist"]
        inputs["pred_conformation"] = node_embedding

        inputs["node_embedding"] = node_embedding
        inputs["distance"] = D_cache


        D_cache, D_M = torch.cdist(conformer_cache, conformer_cache).detach(), make_cdist_mask(node_mask)

        dist_matrix = get_soft_label(torch.cdist(conformer_cache, conformer_cache))
        # dist_matrix = torch.sigmoid(node_embedding @ node_embedding.transpose(1,2))
        # dist_matrix = D_cache
        B, N = dist_matrix.shape[:2]
        mask = torch.eye(N, device=dist_matrix.device).bool().unsqueeze(0).expand(B, -1, -1)
        dist_matrix = dist_matrix.masked_fill(mask, 0)
        dist_matrix = dist_matrix.masked_fill(~D_M.bool(), 0)
        dist_matrix = dist_matrix.masked_fill(inputs["adjacency"].bool(), 1)
        inputs["pred_dist"] = dist_matrix
        inputs["distance"] = dist_matrix
        # inputs["adjacency"] = inputs["adjacency"] * dist_matrix

        sigma, epsilon = inputs.get("sigma"), inputs.get("epsilon")

        # LJ_pot = 4 * epsilon * ((sigma / D_cache) ** 12 - (sigma / D_cache) ** 6)
        s = sigma / D_cache
        s6 = s ** 6

        LJ_force_orig = 24 * epsilon * (s6 / D_cache) * (2 * s6 - 1)
        LJ_force = torch.abs(LJ_force_orig)

        adj_mask = inputs["adjacency"].bool()
        self_mask = torch.eye(D_cache.shape[1], device=D_cache.device).bool().unsqueeze(0).expand(D_cache.shape[0], -1, -1)
        LJ_force[adj_mask | self_mask] = 0
        LJ_force[~D_M.bool()] = 0

        num_retain_edges = inputs.get("num_near_edges")  # (B, N)
        LJ_force = retain_top_k(LJ_force, num_retain_edges, descending=True)
        nonzeros = LJ_force.nonzero(as_tuple=False)
        LJ_force[nonzeros[:, 0], nonzeros[:, 1], nonzeros[:, 2]] = 1

        LJ_nonzero_mask = LJ_force > 0
        LJ_above_zero_mask = LJ_force_orig < 0
        LJ_below_zero_mask = LJ_force_orig > 0

        # LJ_force[LJ_above_zero_mask & LJ_nonzero_mask] = LJ_force[LJ_above_zero_mask & LJ_nonzero_mask]
        LJ_force[LJ_below_zero_mask & LJ_nonzero_mask] = -1
        LJ_force[LJ_above_zero_mask & LJ_nonzero_mask] = 1

        inputs["force_adjacency"] = LJ_force
        del (adj_mask, nonzeros, self_mask, LJ_force, LJ_force_orig, LJ_nonzero_mask,
             LJ_above_zero_mask, LJ_below_zero_mask)  # LJ_attractive, LJ_repulsive, _LJ_force)
        torch.cuda.empty_cache()

        decoder_out = self.decoder(**inputs)
        node_embedding = decoder_out["node_embedding"] #, decoder_out["attn_weight_dict"]

        outputs = self.residual_head(conformer=conformer, hidden_X=node_embedding, padding_mask=node_mask,
                                     compute_loss=True, conformer_base=inputs["pred_conformation"],
                                     adj_t=inputs["adjacency"], near_edge_ratio=inputs.get("near_edge_ratio"))
        return ConformerPredictionOutput(
            loss=(outputs["loss"] + loss_cache) / 2,
            cdist_mae=outputs["cdist_mae"],
            cdist_mse=outputs["cdist_mse"],
            coord_rmsd=outputs["coord_rmsd"],
            conformer=outputs["conformer"],
            conformer_hat=outputs["conformer_hat"],
        )


